const _ = require("lodash");
const path = require("path");
const { extension } = require("mime-types");
const logging = require("@strapi/logger");
const logger = logging.createLogger();
const utils = require("@strapi/utils");
const { ApplicationError, UnauthorizedError, ForbiddenError } = utils.errors;
const jwt = require("jsonwebtoken");

/**
 * Image manipulation functions
 */
const fs = require("fs");
const { join } = require("path");
const sharp = require("sharp");
const { bytesToKbytes } = utils.file;
const THUMBNAIL_RESIZE_OPTIONS = {
  width: 245,
  height: 156,
  fit: "inside",
};

const PUBLIC_FOLDER_NAME = "public";
const UPLOADS_FOLDER_NAME = "uploads";
const PRIVATE = "private";

const writeStreamToFile = (stream, path) =>
  new Promise((resolve, reject) => {
    const writeStream = fs.createWriteStream(path);
    // Reject promise if there is an error with the provided stream
    stream.on("error", reject);
    stream.pipe(writeStream);
    writeStream.on("close", resolve);
    writeStream.on("error", reject);
  });

const getMetadata = (file) =>
  new Promise((resolve, reject) => {
    const pipeline = sharp();
    pipeline.metadata().then(resolve).catch(reject);
    file.getStream().pipe(pipeline);
  });

exports.resizeFileToOverride = async (file, options, { name, hash }) => {
  logger.info("extensions::overrides::resizeFileToOVerride");
  const filePath = join(file.tmpWorkingDirectory, hash);

  await writeStreamToFile(
    file.getStream().pipe(sharp().resize(options)),
    filePath
  );
  const newFile = {
    name,
    hash,
    ext: file.ext,
    mime: file.mime,
    path: file.path || null,
    ownerId: file.ownerId,
    type: file.type,
    getStream: () => fs.createReadStream(filePath),
  };

  const { width, height, size } = await getMetadata(newFile);

  Object.assign(newFile, { width, height, size: bytesToKbytes(size) });
  return newFile;
};

exports.generateThumbnailOverride = async (file) => {
  if (
    file.width > THUMBNAIL_RESIZE_OPTIONS.width ||
    file.height > THUMBNAIL_RESIZE_OPTIONS.height
  ) {
    const newFile = await this.resizeFileToOverride(
      file,
      THUMBNAIL_RESIZE_OPTIONS,
      {
        name: `thumbnail_${file.name}`,
        hash: `thumbnail_${file.hash}`,
      }
    );

    if (file.type) {
      newFile.type = file.type;
    }
    if (file.ownerId) {
      newFile.ownerId = file.ownerId;
    }
    return newFile;
  }

  return null;
};

const crypto = require("crypto");
const { nameToSlug } = require("@strapi/utils");
const { createContext } = require("vm");

const randomSuffix = () => crypto.randomBytes(5).toString("hex");

const generateFileName = (name) => {
  const baseName = nameToSlug(name, { separator: "_", lowercase: false });

  return `${baseName}_${randomSuffix()}`;
};

/**
 * Overriding formatFileInfo function for upload plugin to
 * maintain original file naming.
 */
exports.formatFileInfoOverride = async (
  { filename, type, size },
  fileInfo = {},
  metas = {}
) => {
  const fileService = strapi.plugin("upload").service("file");

  let ext = path.extname(filename);

  if (!ext) {
    ext = `.${extension(type)}`;
  }
  const usedName = (fileInfo.name || filename).normalize();
  const basename = path.basename(usedName, ext);
  console.log("extension::override::fileInfo ", fileInfo);
  console.log("extension::override::type ", type);
  console.log("extension::override::metas ", metas);

  const ctx = strapi.requestContext.get();

  // map through folders
  // set parentfolder to newfolder if it does no exist

  let folderId = await createMediaFolder(metas.type, metas.ownerId);
  console.log("folderId", folderId);
  console.log("folderPath", await fileService.getFolderPath(folderId));
  folderId = folderId ? folderId : fileInfo.folder;

  const entity = {
    ext,
    mime: type,
    hash: generateFileName(basename),
    name: usedName,
    folder: folderId,
    caption: fileInfo.caption,
    alternativeText: fileInfo.alternativeText,
    size: Math.round((size / 1000) * 100) / 100,
    folderPath: await fileService.getFolderPath(folderId),
  };
  console.log("entity ", entity);
  console.log("extension::override::metas ", metas);

  const { refId, ref, field } = metas;

  if (refId && ref && field) {
    entity.related = [
      {
        id: refId,
        __type: ref,
        __pivot: { field },
      },
    ];
  }

  if (metas.path) {
    entity.path = metas.path;
  }

  const { privateName, ownerId } = await getType(fileInfo);

  if (metas.type) {
    entity.type = metas.type;
  } else {
    entity.type = privateName;
  }
  if (metas.ownerId) {
    entity.ownerId = metas.ownerId;
  } else {
    entity.ownerId = ownerId;
  }

  if (metas.tmpWorkingDirectory) {
    entity.tmpWorkingDirectory = metas.tmpWorkingDirectory;
  }

  return entity;
};

const fse = require("fs-extra");
const os = require("os");
const { NotFoundError } = utils.errors;

const createAndAssignTmpWorkingDirectoryToFiles = async (files) => {
  const tmpWorkingDirectory = await fse.mkdtemp(
    path.join(os.tmpdir(), "strapi-upload-")
  );

  if (Array.isArray(files)) {
    files.forEach((file) => {
      file.tmpWorkingDirectory = tmpWorkingDirectory;
    });
  } else {
    files.tmpWorkingDirectory = tmpWorkingDirectory;
  }

  return tmpWorkingDirectory;
};

exports.replaceOverride = async (id, { data, file }, { user } = {}) => {
  const { isImage } = strapi.plugin("upload").service("image-manipulation");
  const uploadService = strapi.plugin("upload").service("upload");

  const config = strapi.config.get("plugin.upload");

  const dbFile = await uploadService.findOne(id);
  if (!dbFile) {
    throw new NotFoundError();
  }

  // create temporary folder to store files for stream manipulation
  const tmpWorkingDirectory = await createAndAssignTmpWorkingDirectoryToFiles(
    file
  );

  let fileData;

  const type = dbFile.url.split("/")[1];
  const ownerId = dbFile.url.split("/")[3];

  try {
    const { fileInfo } = data;
    fileData = await uploadService.enhanceAndValidateFile(file, fileInfo);

    // keep a constant hash and extension so the file url doesn't change when the file is replaced
    _.assign(fileData, {
      hash: dbFile.hash,
      ext: dbFile.ext,
      ownerId: ownerId,
      type: type,
    });

    // execute delete function of the provider
    if (dbFile.provider === config.provider) {
      await strapi.plugin("upload").provider.delete(dbFile);

      if (dbFile.formats) {
        await Promise.all(
          Object.keys(dbFile.formats).map((key) => {
            return strapi.plugin("upload").provider.delete(dbFile.formats[key]);
          })
        );
      }
    }

    // clear old formats
    _.set(fileData, "formats", {});

    if (await isImage(fileData)) {
      await uploadService.uploadImage(fileData);
    } else {
      await strapi.plugin("upload").service("provider").upload(fileData);
    }

    _.set(fileData, "provider", config.provider);
  } finally {
    // delete temporary folder
    await fse.remove(tmpWorkingDirectory);
  }

  return uploadService.update(id, fileData, { user });
};

//path: private/uploads/ownerId/year/fileName
const checkIfFolderExist = async (folderName, parent, user) => {
  const folderService = strapi.plugins.upload.services.folder;
  let folder;
  if (folderName == PRIVATE) {
    folder = await strapi
      .query("plugin::upload.folder")
      .findOne({ where: { name: folderName } });
  } else {
    folder = await strapi.db.query("plugin::upload.folder").findMany({
      where: {
        $and: [{ name: folderName }, { path: { $contains: parent.path } }],
      },
    });
    folder = folder[0];
  }
  if (folder) {
    return folder;
  } else if (parent) {
    console.log("create Folder with parent");
    folder = await folderService.create({
      name: folderName,
      parent: parent.id,
      user: user,
    });
  } else {
    console.log("create Folder");

    folder = await folderService.create({
      name: folderName,
      user: user,
    });
  }
  console.log("folder", folder);

  return folder;
};

const createMediaFolder = async (type, ownerId) => {
  console.log("in create Media Folder");
  console.log("ownerId ", ownerId);
  console.log("type ", type);
  const currentYear = new Date().getFullYear().toString();
  //path: private/uploads/ownerId/year/fileName
  if (type === PRIVATE) {
    const privateFolder = await checkIfFolderExist(PRIVATE, null);
    const uploadsFolder = await checkIfFolderExist(
      UPLOADS_FOLDER_NAME,
      privateFolder
    );
    const ownerIdFolder = await checkIfFolderExist(ownerId, uploadsFolder);
    const currentYearFolder = await checkIfFolderExist(
      currentYear,
      ownerIdFolder
    );
    return currentYearFolder.id;
  }
};

const getType = async (fileInfo) => {
  console.log("in get type");
  const fileService = strapi.plugin("upload").service("file");
  const path = await fileService.getFolderPath(fileInfo.folder);

  console.log("path", path);

  const folderPath = path.split("/");

  const type = await strapi.db.query("plugin::upload.folder").findOne({
    where: {
      path_id: folderPath[1],
    },
  });
  let ownerId;
  if (type.name == "private") {
    ownerId = await strapi.db.query("plugin::upload.folder").findOne({
      where: {
        path_id: folderPath[3],
      },
    });
  }
  console.log("private", type);

  console.log("ownerId", ownerId);

  return { privateName: type.name, ownerId: ownerId?.name };
};
